#![deny(
    missing_docs,
    missing_debug_implementations,
    missing_copy_implementations,
    trivial_casts,
    trivial_numeric_casts,
    unsafe_code,
    unstable_features,
    unused_import_braces,
    unused_qualifications
)]

//!
//! libFenrir is the official rust library implementing the Fenrir protocol

mod config;
pub mod connection;
pub mod dnssec;
pub mod enc;

use ::arc_swap::{ArcSwap, ArcSwapAny};
use ::std::{net::SocketAddr, sync::Arc};
use ::tokio::{net::UdpSocket, task::JoinHandle};

use crate::enc::{asym, sym::CipherKind};
pub use config::Config;
use connection::handshake::{self, Handshake, HandshakeKey};

/// Main fenrir library errors
#[derive(::thiserror::Error, Debug)]
pub enum Error {
    /// The library was not initialized (run .start())
    #[error("not initialized")]
    NotInitialized,
    /// General I/O error
    #[error("IO: {0:?}")]
    IO(#[from] ::std::io::Error),
    /// Dnssec errors
    #[error("Dnssec: {0:?}")]
    Dnssec(#[from] dnssec::Error),
    /// Handshake errors
    #[error("Handshake: {0:?}")]
    Handshake(#[from] handshake::Error),
    /// Key error
    #[error("key: {0:?}")]
    Key(#[from] crate::enc::Error),
}

struct FenrirInner {
    key_exchanges: ArcSwapAny<Arc<Vec<(asym::Key, asym::KeyExchange)>>>,
    ciphers: ArcSwapAny<Arc<Vec<CipherKind>>>,
    keys: ArcSwapAny<Arc<Vec<HandshakeKey>>>,
}

impl FenrirInner {
    fn recv_handshake(&self, handshake: Handshake) -> Result<(), Error> {
        use connection::handshake::{dirsync::DirSync, HandshakeData};
        match handshake.data {
            HandshakeData::DirSync(ds) => match ds {
                DirSync::Req(req) => {
                    let ephemeral_key = {
                        // Keep this block short to avoid contention
                        // on self.keys
                        let keys = self.keys.load();
                        if let Some(h_k) =
                            keys.iter().find(|k| k.id == req.key_id)
                        {
                            use enc::asym::PrivKey;
                            // Directory synchronized can only used keys
                            // for key exchange, not signing keys
                            if let PrivKey::Exchange(k) = &h_k.key {
                                Some(k.clone())
                            } else {
                                None
                            }
                        } else {
                            None
                        }
                    };
                    if ephemeral_key.is_none() {
                        ::tracing::debug!("No such key id: {:?}", req.key_id);
                        return Err(handshake::Error::UnknownKeyID.into());
                    }
                    let ephemeral_key = ephemeral_key.unwrap();
                    {
                        let exchanges = self.key_exchanges.load();
                        if None
                            == exchanges.iter().find(|&x| {
                                *x == (ephemeral_key.kind(), req.exchange)
                            })
                        {
                            return Err(
                                enc::Error::UnsupportedKeyExchange.into()
                            );
                        }
                    }
                    {
                        let ciphers = self.ciphers.load();
                        if None == ciphers.iter().find(|&x| *x == req.cipher) {
                            return Err(enc::Error::UnsupportedCipher.into());
                        }
                    }
                    let shared_key = match ephemeral_key
                        .key_exchange(req.exchange, req.exchange_key)
                    {
                        Ok(shared_key) => shared_key,
                        Err(e) => return Err(handshake::Error::Key(e).into()),
                    };
                    todo!();
                }
                DirSync::Resp(resp) => {
                    todo!();
                }
            },
        }
    }
}

/// Instance of a fenrir endpoint
#[allow(missing_copy_implementations, missing_debug_implementations)]
pub struct Fenrir {
    /// library Configuration
    cfg: Config,
    /// listening udp sockets
    sockets: Vec<(Arc<UdpSocket>, JoinHandle<::std::io::Result<()>>)>,
    /// DNSSEC resolver, with failovers
    dnssec: Option<dnssec::Dnssec>,
    /// Broadcast channel to tell workers to stop working
    stop_working: ::tokio::sync::broadcast::Sender<bool>,
    /// Private keys used in the handshake
    _inner: Arc<FenrirInner>,
}

// TODO: graceful vs immediate stop

impl Drop for Fenrir {
    fn drop(&mut self) {
        self.stop_sync()
    }
}

impl Fenrir {
    /// Create a new Fenrir endpoint
    pub fn new(config: &Config) -> Result<Self, Error> {
        let listen_num = config.listen.len();
        let (sender, _) = ::tokio::sync::broadcast::channel(1);
        let endpoint = Fenrir {
            cfg: config.clone(),
            sockets: Vec::with_capacity(listen_num),
            dnssec: None,
            stop_working: sender,
            _inner: Arc::new(FenrirInner {
                ciphers: ArcSwapAny::new(Arc::new(Vec::new())),
                key_exchanges: ArcSwapAny::new(Arc::new(Vec::new())),
                keys: ArcSwapAny::new(Arc::new(Vec::new())),
            }),
        };
        Ok(endpoint)
    }

    /// Start all workers, listeners
    pub async fn start(&mut self) -> Result<(), Error> {
        if let Err(e) = self.add_sockets().await {
            self.stop().await;
            return Err(e.into());
        }
        self.dnssec = Some(dnssec::Dnssec::new(&self.cfg.resolvers).await?);
        Ok(())
    }

    /// Stop all workers, listeners
    /// asyncronous version for Drop
    fn stop_sync(&mut self) {
        let _ = self.stop_working.send(true);
        let mut toempty_socket = Vec::new();
        ::std::mem::swap(&mut self.sockets, &mut toempty_socket);
        let task = ::tokio::task::spawn(Self::stop_sockets(toempty_socket));
        let _ = ::futures::executor::block_on(task);
        self.dnssec = None;
    }

    /// Stop all workers, listeners
    pub async fn stop(&mut self) {
        let _ = self.stop_working.send(true);
        let mut toempty_socket = Vec::new();
        ::std::mem::swap(&mut self.sockets, &mut toempty_socket);
        Self::stop_sockets(toempty_socket).await;
        self.dnssec = None;
    }

    /// actually do the work of stopping resolvers and listeners
    async fn stop_sockets(
        sockets: Vec<(Arc<UdpSocket>, JoinHandle<::std::io::Result<()>>)>,
    ) {
        for s in sockets.into_iter() {
            let _ = s.1.await;
        }
    }

    /// Enable some common socket options. This is just the unsafe part
    fn enable_sock_opt(
        fd: ::std::os::fd::RawFd,
        option: ::libc::c_int,
        value: ::libc::c_int,
    ) -> ::std::io::Result<()> {
        #[allow(unsafe_code)]
        unsafe {
            #[allow(trivial_casts)]
            let val = &value as *const _ as *const ::libc::c_void;
            let size = ::std::mem::size_of_val(&value) as ::libc::socklen_t;
            // always clear the error bit before doing a new syscall
            let _ = ::std::io::Error::last_os_error();
            let ret =
                ::libc::setsockopt(fd, ::libc::SOL_SOCKET, option, val, size);
            if ret != 0 {
                return Err(::std::io::Error::last_os_error());
            }
        }
        Ok(())
    }
    /// Add an async udp listener
    async fn bind_udp(sock: SocketAddr) -> ::std::io::Result<UdpSocket> {
        let socket = UdpSocket::bind(sock).await?;

        use ::std::os::fd::AsRawFd;
        let fd = socket.as_raw_fd();
        // can be useful later on for reloads
        Self::enable_sock_opt(fd, ::libc::SO_REUSEADDR, 1)?;
        Self::enable_sock_opt(fd, ::libc::SO_REUSEPORT, 1)?;

        // We will do path MTU discovery by ourselves,
        // always set the "don't fragment" bit
        if sock.is_ipv6() {
            Self::enable_sock_opt(fd, ::libc::IPV6_DONTFRAG, 1)?;
        } else {
            // FIXME: linux only
            Self::enable_sock_opt(
                fd,
                ::libc::IP_MTU_DISCOVER,
                ::libc::IP_PMTUDISC_DO,
            )?;
        }

        Ok(socket)
    }

    /// Run a dedicated loop to read packets on the listening socket
    async fn listen_udp(
        mut stop_working: ::tokio::sync::broadcast::Receiver<bool>,
        fenrir: Arc<FenrirInner>,
        socket: Arc<UdpSocket>,
    ) -> ::std::io::Result<()> {
        // jumbo frames are 9K max
        let mut buffer: [u8; 9000] = [0; 9000];
        loop {
            let (bytes, sock_from) = ::tokio::select! {
                _done = stop_working.recv() => {
                    break;
                }
                result = socket.recv_from(&mut buffer) => {
                    result?
                }
            };
            Self::recv(fenrir.clone(), &buffer[0..bytes], sock_from).await;
        }
        Ok(())
    }

    /// Add all UDP sockets found in config
    /// and start listening for packets
    async fn add_sockets(&mut self) -> ::std::io::Result<()> {
        let sockets = self.cfg.listen.iter().map(|s_addr| async {
            let socket =
                ::tokio::spawn(Self::bind_udp(s_addr.clone())).await??;
            Ok(Arc::new(socket))
        });
        let sockets = ::futures::future::join_all(sockets).await;
        for s_res in sockets.into_iter() {
            match s_res {
                Ok(s) => {
                    let stop_working = self.stop_working.subscribe();
                    let join = ::tokio::spawn(Self::listen_udp(
                        stop_working,
                        self._inner.clone(),
                        s.clone(),
                    ));
                    self.sockets.push((s, join));
                }
                Err(e) => {
                    return Err(e);
                }
            }
        }
        Ok(())
    }

    /// Get the raw TXT record of a Fenrir domain
    pub async fn resolv_str(&self, domain: &str) -> Result<String, Error> {
        match &self.dnssec {
            Some(dnssec) => Ok(dnssec.resolv(domain).await?),
            None => Err(Error::NotInitialized),
        }
    }

    /// Get the raw TXT record of a Fenrir domain
    pub async fn resolv(&self, domain: &str) -> Result<dnssec::Record, Error> {
        let record_str = self.resolv_str(domain).await?;
        Ok(dnssec::Dnssec::parse_txt_record(&record_str)?)
    }

    const MIN_PACKET_BYTES: usize = 8;
    /// Read and do stuff with the udp packet
    async fn recv(
        fenrir: Arc<FenrirInner>,
        buffer: &[u8],
        _sock_from: SocketAddr,
    ) {
        if buffer.len() < Self::MIN_PACKET_BYTES {
            return;
        }
        use connection::ID;
        let raw_id: [u8; 8] = buffer.try_into().expect("unreachable");
        if ID::from(raw_id).is_handshake() {
            use connection::handshake::Handshake;
            let handshake = match Handshake::parse(&buffer[8..]) {
                Ok(handshake) => handshake,
                Err(e) => {
                    ::tracing::warn!("Handshake parsing: {}", e);
                    return;
                }
            };
            if let Err(err) = fenrir.recv_handshake(handshake) {
                ::tracing::debug!("Handshake recv error {}", err);
                return;
            }
        }
        // copy packet, spawn
        todo!();
    }
}
